
/*
 * @(#)MultiPanelLayout.java	1.00 98/12/28
 * 
 * Modification Log:
 *
 */

package mplayout;

import java.awt.*;
import java.util.Vector;


/**
 * Lays out multiple panels containing components within a parent container.
 * Maintains left and right alignment of all components in all panels.
 * <br>
 * This is the only layout that operates on multiple containers at once.
 * The parent container has to be designated as the master. Calling layout
 * on this container lays out all other dependent containers. Calling
 * layout on a dependent container has no effect.
 * <p>
 * Steps to use:
 * <li>Decide the screen grid size, lets say 80 by 24.
 * <li>Create the parent container (frame, lets say).
 * <li>Create one instance of MultiPanelLayout with the above mentioned parameters.
 * <li>Set layout manager of the parent container to this class.
 * <li>Lets say you have five panels to display in the frame, p1 to p5. The panels 
 *     may overlap each other.
 * <li>Decide the grid dimensions of the containers and of the components to be added
 *     within them.
 * <li>Add the dependent containers (panels p1 to p5) to the parent container.
 * <li>Add the components to their respective panels.
 * <li>Note that a new instance of the constraint class MultiPanelLayoutConstraint has
 *     to be created for every component added. Unlike GridBagLayout, a new copy of the
 *     class is NOT created internally.
 * <p>
 *
 * You are free to use this code and to make modifications provided
 * this notice is retained.
 * <p>
 * If you found this useful, please add a note of acknowledgement to my guestbook,
 * or send me a mail at <a href="mailto:tanmaykm@hotmail.com">tanmaykm@hotmail.com</a>. If you would like to report a bug or 
 * suggest some improvements, you are most welcome. I will be happy to help you use
 * this piece of code.
 * <p>
 * @author 	    Tanmay K. Mohapatra
 * @version     1.00, 28 Dec 1998
 */

public class MultiPanelLayout implements java.awt.LayoutManager2
{
    /** the maximum grid sizes */
    private int                 iMaxX, iMaxY;

    /** containers using this layout */
    private Vector              containerVect;
    /** components added to this layout */
    private Vector              componentVect;
    /** constraints for all components added to this layout */
    private Vector              constraintVect;

    /** the parent container which triggers relayout of all others */
    private Container           parentContainer;

    /** the class that calculates the positions */
    private MPConstraintMatrix  matrix;

    /** the insets of the parent container */
    private Insets              parentInsets;

    /** container list used to determine their visibility */
    private Component           []allCont;
    /** visibility indicators for each container */
    private boolean             []allVisib;

    /**
     * Constructor. The layout must be constructed only once and the same
     * instance must be used in all containers which are to be laid out
     * together.
     * @param   iMaxX   the number of grids in the horizontal direction
     * @param   iMaxY   the number of grids in the vertical direction
     * @param   parentContainer the parent container containing all components and other containers
     */
    public MultiPanelLayout(int iMaxX, int iMaxY, Container parentContainer)
    {
        this.iMaxX = iMaxX;
        this.iMaxY = iMaxY;
        this.parentContainer = parentContainer;
        matrix = new MPConstraintMatrix(iMaxX, iMaxY);
        containerVect = new Vector(10, 3);
        componentVect = new Vector(30, 20);
        constraintVect = new Vector(30, 20);
    }

    /**
     * Adds a component to the layout, using a constraint object. 
     * Determines its parent container and adds it to the list of parents.
     */
    public void addLayoutComponent(Component comp, Object constraints)
    {
        Container parent;

        componentVect.addElement(comp);
        constraintVect.addElement(constraints);
        if ( (comp instanceof Container) && 
             (((Container)comp).getComponentCount() > 0) &&
             componentVect.contains((((Container)comp).getComponent(0))) &&
             (!containerVect.contains(comp)) )
        {
            // add to list of parents if some component inside this is already there
            containerVect.addElement(comp);
        }
        parent = comp.getParent();
        if ((parent != parentContainer) && (!containerVect.contains(parent)))
        {
            containerVect.addElement(parent);
        }
    }

    /**
     * Returns the alignment along the x axis. This specifies how the component 
     * would like to be aligned relative to other components. The value should 
     * be a number between 0 and 1 where 0 represents alignment along the 
     * origin, 1 is aligned the furthest away from the origin, 0.5 is centered, 
     * etc.
     */
    public float getLayoutAlignmentX(Container target)
    {
        return 0.5f;
    }

    /**
     * Returns the alignment along the y axis. This specifies how the component 
     * would like to be aligned relative to other components. The value should 
     * be a number between 0 and 1 where 0 represents alignment along the 
     * origin, 1 is aligned the furthest away from the origin, 0.5 is centered, 
     * etc.
     */
    public float getLayoutAlignmentY(Container target)
    {
        return 0.5f;
    }

    /**
     * Invalidates the layout, indicating that if the layout manager has 
     * cached information it should be discarded.
     */
    public void invalidateLayout(Container target)
    {
    }

    /**
     * Adds the specified component with the specified name to the layout.
     */
    public void addLayoutComponent(String name, Component comp)
    {
        // not to be coded
    }

    /**
     * Lays out the container in the specified panel.
     */
    public void layoutContainer(Container parent)
    {
        int                         iIndex, iNumComps;
        int                         iCompIndex;
        MultiPanelLayoutConstraint  compShape;
        Dimension                   dimParent;

        //System.out.println("layout start...");
        if (parent != parentContainer)
        {
            //System.out.println("layout not reqd...");
            return;
        }

        calculateGrid();
        dimParent = parentContainer.getSize();
        parentInsets = parentContainer.getInsets();
        matrix.fitMatrixTo(dimParent.width-parentInsets.right-parentInsets.left, 
                           dimParent.height-parentInsets.top-parentInsets.bottom);

        // go thru all components and reshape them
        iNumComps = containerVect.size();

        // first reshape the containers
        for (iIndex=0; iIndex<iNumComps; iIndex++)
        {
            iCompIndex = componentVect.indexOf(containerVect.elementAt(iIndex));
            reshapeCompAt(iCompIndex, parentInsets);
        }

        // go thru all components and reshape them
        iNumComps = componentVect.size();

        // go thru all components
        for (iIndex=0; iIndex<iNumComps; iIndex++)
        {
            Component comp;
            if (!containerVect.contains(comp=(Component)componentVect.elementAt(iIndex)) &&
                comp.isShowing() && isContVisible(comp.getParent()) )
            {
                reshapeCompAt(iIndex, null);
            }
        }
        //System.out.println("layout end...");
    }


    /**
     * reshape the component at position iIndex of the component vector
     * @param   iIndex  the index of the component
     * @param   insets  the insets of the parent container for this component
     */
    private void reshapeCompAt(int iIndex, Insets insets)
    {
        MultiPanelLayoutConstraint      compShape;
        Component                       comp;
        Rectangle                       rect;
        int                             iStartX, iStartY, iEndX, iEndY;
        int                             iParentStart, iCompStart;

        // get the component and its constraints
        comp = (Component)componentVect.elementAt(iIndex);
        compShape = (MultiPanelLayoutConstraint)constraintVect.elementAt(iIndex);
        rect = new Rectangle();

        // calculate height
        iStartY = compShape.iParentStartY+compShape.iStartY;
        iEndY = compShape.iParentStartY+compShape.iStartY+compShape.iHeight;
        rect.height = matrix.aiVertSizes[iEndY-1];
        if (iStartY > 0) // start y will be -1 for the top most component
        {
            rect.height -= matrix.aiVertSizes[iStartY-1];
        }

        // calculate width
        iStartX = compShape.iParentStartX+compShape.iStartX;
        iEndX = compShape.iParentStartX+compShape.iStartX+compShape.iWidth;
        rect.width  = matrix.aiHorzSizes[iEndX-1];
        if(iStartX > 0) // start x will be -1 for the left most component
        {
            rect.width -= matrix.aiHorzSizes[iStartX-1];
        }

        // calculate the horz start point
        iParentStart = compShape.iParentStartX;
        iCompStart = compShape.iParentStartX+compShape.iStartX;
        rect.x = 0;
        if (iCompStart > 0) // will be -1 for the left most component
        {
            rect.x += matrix.aiHorzSizes[iCompStart-1];
        }
        if (iParentStart > 0) // will be -1 for the left most parent
        {
            rect.x -= matrix.aiHorzSizes[iParentStart-1];
        }
        // calculate the vert start point
        iParentStart = compShape.iParentStartY;
        iCompStart = compShape.iParentStartY+compShape.iStartY;
        rect.y = 0;
        if (iCompStart > 0) // will be -1 for the left most component
        {
            rect.y += matrix.aiVertSizes[iCompStart-1];
        }
        if (iParentStart > 0) // will be -1 for the left most parent
        {
            rect.y -= matrix.aiVertSizes[iParentStart-1];
        }

        // take care of the insets
        if (null != insets)
        {
            rect.x += insets.left;
            rect.y += insets.top;
        }

        comp.setBounds(rect);

        //System.out.println("compShape ParentX=" + compShape.iParentStartX +
        //    " ParentY=" + compShape.iParentStartY +
        //    " StartX=" + compShape.iStartX +
        //    " StartY=" + compShape.iStartY +
        //    " width=" + compShape.iWidth +
        //    " height=" + compShape.iHeight);
        //System.out.println("Comp at " + iIndex + " reshaped to " + rect);
    }

    /**
     * Returns the maximum size of this component.
     */
    public Dimension maximumLayoutSize(Container parent)
    {
        calculateGrid();
        if (parent == parentContainer)
        {
            return new Dimension(matrix.aiHorzSizes[iMaxX-1], matrix.aiVertSizes[iMaxY-1]);
        }
        return new Dimension(2, 2);
    }

    /**
     * Calculates the minimum size dimensions for the specified panel given 
     * the components in the specified parent container.
     */
    public Dimension minimumLayoutSize(Container parent)
    {
        calculateGrid();
        if (parent == parentContainer)
        {
            return new Dimension(matrix.aiHorzSizes[iMaxX-1], matrix.aiVertSizes[iMaxY-1]);
        }
        return new Dimension(2, 2);
    }

    /**
     * Calculates the preferred size dimensions for the specified panel given 
     * the components in the specified parent container.
     */
    public Dimension preferredLayoutSize(Container parent)
    {
        calculateGrid();
        if (parent == parentContainer)
        {
            return new Dimension(matrix.aiHorzSizes[iMaxX-1], matrix.aiVertSizes[iMaxY-1]);
        }
        return new Dimension(2, 2);
    }

    /**
     * Removes the specified component from the layout.
     */
    public void removeLayoutComponent(Component comp)
    {
        int iIndex = componentVect.indexOf(comp);

        if (iIndex >= 0)
        {
            componentVect.removeElementAt(iIndex);
            constraintVect.removeElementAt(iIndex);
        }
        if (containerVect.contains(comp))
        {
            containerVect.removeElement(comp);
        }
    }

    /**
     * Calculate the grid positions.
     */
    private void calculateGrid()
    {
        int                             iNumComps, iIndex;
        Component                       comp;
        Dimension                       prefSize;
        MultiPanelLayoutConstraint      compShape;

        matrix.reset();
        calcContVisib();

        iNumComps = componentVect.size();

        // go thru all components
        for (iIndex=0; iIndex<iNumComps; iIndex++)
        {
            comp = (Component)componentVect.elementAt(iIndex);

            if ( (-1 != containerVect.indexOf(comp)) || !comp.isShowing() || !isContVisible(comp.getParent()))
            {
                // don't add containers or components that are not visible
                //System.out.println("Component visibility: " + comp.isShowing());
                continue;
            }

            // find preferredSize
            prefSize = comp.getPreferredSize();
            //System.out.println("Size = " + prefSize);
            // get the constraints
            compShape = (MultiPanelLayoutConstraint)constraintVect.elementAt(iIndex);

            matrix.addConstraint(MPConstraintMatrix.VERT_TYPE, 
                                 compShape.iParentStartY+compShape.iStartY+compShape.iHeight-1,
                                 compShape.iParentStartY+compShape.iStartY-1,
                                 prefSize.height);
            matrix.addConstraint(MPConstraintMatrix.HORZ_TYPE, 
                                 compShape.iParentStartX+compShape.iStartX+compShape.iWidth-1,
                                 compShape.iParentStartX+compShape.iStartX-1,
                                 prefSize.width);
        }

        matrix.solveMatrix();
    }
    
    /**
     * Calculate the visibility of all containers within the parent container.
     * Components within containers that are overlapped completely by other containers
     * need not be added to the layout.
     */
    private void calcContVisib()
    {
        int                             iIndex, iNumConts, iVectIndex, iUnionIndex;
        Rectangle                       unionRect, thisRect, fullRect, allRects[];
        MultiPanelLayoutConstraint      compShape;
        
        allCont = parentContainer.getComponents();
        if(0 < (iNumConts = allCont.length))
        {
            allVisib = new boolean[iNumConts];
            allRects = new Rectangle[iNumConts];
            fullRect = new Rectangle(iMaxX, iMaxY);

            for (iIndex=0; iIndex < iNumConts; iIndex++)
            {
                // if this container is in our container list
                if (containerVect.contains(allCont[iIndex]))
                {
                    // get the constraints
                    compShape = (MultiPanelLayoutConstraint)constraintVect.elementAt(componentVect.indexOf(allCont[iIndex]));
                    allRects[iIndex] = new Rectangle(compShape.iParentStartX+compShape.iStartX, 
                                                     compShape.iParentStartY+compShape.iStartY, 
                                                     compShape.iWidth, 
                                                     compShape.iHeight);
                    // get the union of all rectangles
                    for(iUnionIndex=1, unionRect=allRects[0]; iUnionIndex <= iIndex; iUnionIndex++)
                    {
                        unionRect.add(allRects[iUnionIndex]);
                    }
                    // this is visible
                    allVisib[iIndex] = true;
                    // if the union equals the full size, the rest of 'em are not visible
                    if (unionRect.equals(fullRect))
                    {
                        break;  // booleans are born as false
                    }
                }
            }
        }
    }
 
    /**
     * @return  true    if the container is not hidden by 
     *                  any other container on top of it, false otherwise
     * @param   cont    the container to test
     */
    private boolean isContVisible(Container cont)
    {
        int         iIndex, iNumConts;
        boolean     bRet = false;
        
        for(iIndex=0, iNumConts=allCont.length; iIndex < iNumConts; iIndex++)
        {
            if (allCont[iIndex].equals(cont))
            {
                bRet = allVisib[iIndex];
                break;
            }
        }
        //System.out.println("Cont: " + cont + " visib=" + bRet);
        return bRet;
    }
}

